dofile_once("data/scripts/lib/utilities.lua")
dofile_once("mods/scoita/files/scripts/libs/json.lua")
dofile_once("mods/scoita/files/scripts/utils.lua")
dofile_once("mods/scoita/files/scripts/rules/RuleManager.lua")
dofile_once("mods/scoita/files/scripts/ruleTriggers/RuleTriggerManager.lua")
dofile_once("mods/scoita/files/scripts/rules/knownRuleTypes.lua")
dofile_once("mods/scoita/files/scripts/scoreManager/scoreManagerInterface.lua")


dofile_once("mods/scoita/files/scripts/apiTypes.lua")

EventManager = {}

function EventManager:new(o)
    local this = o or {
      eventBuffer = {},
      ruleManager = RuleManager:new(),
      scoreManager = nil,
    }

    setmetatable(this,self)
    self.__index = self

    
    function this:connectToScoreManager(scoreManager)
      self.scoreManager = scoreManager
    end


    function this:init()
      GlobalsSetValue(KnownGlobals.ScoreEventBuffer,"[]")

      local playerStateTracker = self.scoreManager:getStateTrackerForplayer()
      table.insert(playerStateTracker.trackingHelpers.holyMountainVisitedYCoOrd,-1)
    end


    function this:onModInitialising(ModTextFileSetContentFunc, ModMagicNumbersFileAddFunc)
    end


    function this:initGameBasedOnGameMode(gameModeInstance)
      local playerStateTracker = self.scoreManager:getStateTrackerForplayer()

      self:configureRulesBasedOnGameMode(gameModeInstance)
      self:configureTriggersBasedOnGameMode(gameModeInstance)
      
      if(gameModeInstance.gameMode.modeInfo.config.hasTimeLimit) then
        playerStateTracker.trackingHelpers.runTimer.totalSecondsAllowed = gameModeInstance.gameMode.modeInfo.config.timeLimit
        playerStateTracker.trackingHelpers.runTimer.enabled = true
      end

      local playerEntity = getPlayerEntity()

      -- Add script to player if it doesn't exist (happends when reloading game, esp when dev testing)
      local playerSciptDamageReceivedName = "mods/scoita/files/scripts/dataScriptExtensions/playerDamageReceived.lua"
      local luaComponentsOnPlayer = EntityGetComponent(playerEntity, "LuaComponent")
      local playerSciptDamageReceivedScriptExists = false
      for i,v in pairs(luaComponentsOnPlayer) do
        if(ComponentGetValue2(v,"script_damage_received") == playerSciptDamageReceivedName) then
          playerSciptDamageReceivedScriptExists = true
        end
      end

      if(not playerSciptDamageReceivedScriptExists) then
        EntityAddComponent( playerEntity, "LuaComponent", {
          script_damage_received=playerSciptDamageReceivedName
        });
      end
      

    end


    function this:configureRulesBasedOnGameMode(gameModeInstance)
      
      for idx,ruleConfig in ipairs(gameModeInstance.gameMode.modeInfo.config.rulesConfig.inbuiltRules) do
        
        if(ruleConfig.isEnabled) then
          self.ruleManager:setIsEnabled(ruleConfig.name, true)
          if(ruleConfig.config ~= nil) then
            self.ruleManager:setConfig(ruleConfig.name, true)
          end
        end

      end

      for idx,inbuiltRuleTag in ipairs(gameModeInstance.gameMode.modeInfo.config.rulesConfig.inbuiltRuleTagsToEnable) do
        self.ruleManager:enableTaggedRules(inbuiltRuleTag)
      end

    end


    function this:configureTriggersBasedOnGameMode(gameModeInstance)
      local triggers = gameModeInstance.gameMode.modeInfo.config.rulesConfig.triggers
      local playerState = self.scoreManager:getStateTrackerForplayer()
      local runState = self.scoreManager:getCurrentRun()

      for k,triggerInfo in pairs(triggers) do
        local trigger = ruleTriggerManager:createNewTrigger(triggerInfo.type, triggerInfo)
        local triggerState = trigger:createBaseState(triggerInfo)
        
        runState.ruleHelpers.triggers[trigger.id] = trigger
        playerState.ruleStates.triggerStates[trigger.id] = triggerState
      end

    end


    function this:processBufferedScoreEvents()     
      
      local scoreBufGlobal = GlobalsGetValue(KnownGlobals.ScoreEventBuffer);
      
      if(scoreBufGlobal == nil or scoreBufGlobal == "") then
        --print(KnownGlobals.ScoreEventBuffer..' isnt initialised - Not processing scores')
        do return end
      end

      local buffer = json.decode(scoreBufGlobal)
      
      local bufferLen = tablelength(buffer)
      if (bufferLen == 0) then return end
      
      for i=1,bufferLen do
        local event = buffer[i]
        self:processScoreEventForPlayer(event)
      end

      buffer = {}
      GlobalsSetValue(KnownGlobals.ScoreEventBuffer,"[]")
    end


    function this:setScoreForPlayer(newScore)
      local playerStateTracker = self.scoreManager:getStateTrackerForplayer()
      playerStateTracker.currentScore = newScore
    end


    function this:processScoreEventForPlayer(scoreEvent, doesRunUseScoringSystem)
      local isOnScoitaRun = self.scoreManager:isOnScoitaRun()
      local doesRunUseScoringSystem = isOnScoitaRun and self.scoreManager:getCurrentGameModeInstance().gameMode.modeInfo.config.scoreRulesInfo.calculationType == ScoitaScoreCalculationType.pointsSystem
      local shouldAddPointsToScore = true
      if(isOnScoitaRun and doesRunUseScoringSystem == false) then
        shouldAddPointsToScore = false
      end

      -- print("isOnScoitaRun "..tostring(isOnScoitaRun))
      -- print("doesRunUseScoringSystem "..tostring(doesRunUseScoringSystem))
      -- print("shouldAddPointsToScore "..tostring(shouldAddPointsToScore))
      self:processRuleTriggers(scoreEvent)

      if(shouldAddPointsToScore and scoreEvent.points ~= nil) then
        local playerStateTracker = self.scoreManager:getStateTrackerForplayer()
        playerStateTracker.currentScore = playerStateTracker.currentScore + scoreEvent.points
        -- add timestamp for score event
        scoreEvent.frameNum = GameGetFrameNum()
      
        scoreEvent.context = nil
        scoreEvent.messageType = nil
        table.insert(playerStateTracker.logs,scoreEvent)
      end

    end

    function this:triggerScoitaWin()
      GameAddFlagRun("scoita_run_completed")
      self.scoreManager:superMurderThePlayer()
    end
    

    function this:processRuleTriggers(scoreEvent)
      local runState = self.scoreManager:getCurrentRun()
      local playerState = self.scoreManager:getStateTrackerForplayer()

      for id,trigger in pairs(runState.ruleHelpers.triggers) do
        --print("Name:"..trigger.config.friendlyName)

        if(scoreEvent.metaData ~= nil and scoreEvent.metaData.eventType ~= nil) then
          local shouldProcessRule = trigger.eventsToListen[scoreEvent.metaData.eventType]
          --print("EventType "..scoreEvent.metaData.eventType)
          --print("shouldProcessRule "..tostring(shouldProcessRule))
          --print(json.encode(trigger.eventsToListen))
          
          if(shouldProcessRule) then
            local triggerState = playerState.ruleStates.triggerStates[trigger.id]
            if(triggerState == nil) then
              print("WARNING : NO TRIGGER STATE")
            end
            
            local triggerResult = trigger:processEvent(scoreEvent, triggerState)

            -- LAME FOR NOW !
            if(triggerResult.actionToTrigger == "winTheGame") then
              self:triggerScoitaWin()
            end
            
          end

        end

      end
      
    end


    function this:getLastVisitedHolyMountainCoOrds()
      local playerStateTracker = self.scoreManager:getStateTrackerForplayer()
      return playerStateTracker.trackingHelpers.holyMountainVisitedYCoOrd
    end


    function this:checkScoreEvents()
      self:checkHolyNewMountainReached()        
      self:checkNewDepthReached()

      self:processTimeBasedScoreUpdates()
      self:processUpdateTriggers()
    end

    function this:getTriggersForRun()
      return self.scoreManager:getCurrentRun().ruleHelpers.triggers
    end

    function this:getTriggerStatesForPlayer()
      return self.scoreManager:getStateTrackerForplayer().ruleStates.triggerStates
    end

    function this:processUpdateTriggers()
      local runState = self.scoreManager:getCurrentRun()
      local playerState = self.scoreManager:getStateTrackerForplayer()

      
      -- Careful with this, reusing same state object for each trigger for perf
      local triggerUpdateState = {
        scoreManager = scoreManager,
      }

      for id,trigger in pairs(runState.ruleHelpers.triggers) do
        local shouldProcessTrigger = trigger.shouldCheckEveryFrame
          
        if(shouldProcessTrigger) then
          local triggerState = playerState.ruleStates.triggerStates[trigger.id]
          
          if(triggerState == nil) then
            print("WARNING : NO TRIGGER STATE")
          end

          triggerUpdateState.triggerState = triggerState
          trigger:updateTrigger(triggerUpdateState)

        end
      end

      this:checkWinConditions(runState.ruleHelpers.triggers, playerState.ruleStates.triggerStates)
    end

    function this:checkWinConditions(triggers, triggerStates)
      local winConditionTriggers = {}
      local winConditionsMet = 0
      for id,trigger in pairs(triggers) do
        if(trigger:hasTag(ScoitaRuleTriggerTags.winCondition)) then
          hasWinConditions = true
          local triggerState = triggerStates[trigger.id]
          
          if(triggerState.state.isTriggered == true) then
            winConditionsMet = winConditionsMet + 1
          end

          if(triggerState.state.isFailed == true) then
            self.scoreManager:superMurderThePlayer()
          end

          table.insert(winConditionTriggers, trigger)
        end
      end

      if(hasWinConditions) then
          if(winConditionsMet == tablelength(winConditionTriggers)) then
            self:triggerScoitaWin()
          end
      end
    
    end

    function this:processTimeBasedScoreUpdates()
      local playerStateTracker = self.scoreManager:getStateTrackerForplayer()
      local isOnScoitaRun = self.scoreManager:isOnScoitaRun()

      if(isOnScoitaRun) then
        local instance = self.scoreManager:getCurrentGameModeInstance()
        local isRunActive = self.scoreManager:isRunActive()
        local isTimeBasedScoring = instance.gameMode.modeInfo.config.scoreRulesInfo.typeMeta == ScoitaScoreTypeMeta.time

        if(isRunActive and isTimeBasedScoring) then
          self.scoreManager:setScoreForPlayer(self.scoreManager:getActiveRunDurationInSeconds())
        end
      end

      if(playerStateTracker.trackingHelpers.runTimer.enabled) then
        local secondsRemaining = self.scoreManager:getSecondsRemainingInRun()

        if(secondsRemaining <= 0) then
          self.scoreManager:superMurderThePlayer()
          --GameTriggerGameOver()
        end
      end
    end

    function this:checkHolyNewMountainReached()
      local curBiome = getPlayersCurrentBiome()

      if(curBiome == "$biome_holymountain") then
        local player = EntityGetWithTag("player_unit")[1]
        local x, y = EntityGetTransform(player)
        local hmYCoOrds = self:getLastVisitedHolyMountainCoOrds()
        local holyMountainVisitedCount = tablelength(hmYCoOrds)
        
        local largestHolyMountainY = -1
        for k,v in pairs(hmYCoOrds) do
          if(largestHolyMountainY < v) then
            largestHolyMountainY = v
          end
        end

        if(y > (largestHolyMountainY + 1000)) then
          table.insert(hmYCoOrds,y)
          local sce = {desc = "Reached Holy Mountain x"..holyMountainVisitedCount, points = 100 * holyMountainVisitedCount}
          addScoreEventForPlayer(sce)
        end
      end
    end

    function this:checkNewDepthReached()
      if(self.ruleManager:isRuleEnabled(KnownRuleNames.OnReachNewDepth)) then
        local player = EntityGetWithTag("player_unit")[1]
        local x, y = EntityGetTransform(player)
        local curDepth = math.floor(y) / 10
        if(self.scoreManager:getScoreForPlayer() < curDepth) then
          self.scoreManager:setScoreForPlayer(curDepth)
        end
      end
    end

    return this
end

EventManager = EventManager:new()