dofile_once("mods/scoita/files/scripts/rules/knownRuleTypes.lua")

local onBasicEnemyKillRule = {
    name = KnownRuleNames.OnBasicEnemyKill,
    desc = "Killing enemies awards points based on health",
    tags = "standardPointsRuleset,",
    isEnabled = false,
    scale = 1,
    config = {},
    processEvent = "return function processEvent(context,event) ****do a simple score cala here***** end"
}

local onReachEachHolyMountainRule = {
    name = KnownRuleNames.OnReachEachHolyMountain,
    desc = "Reaching each new holy mountain awards increasing points",
    tags = "standardPointsRuleset,",
    isEnabled = false,
    scale = 1,
    config = {},
}

local onBossKillRule = {
    name = KnownRuleNames.OnBossKill,
    desc = "Killing bosses",
    tags = "standardPointsRuleset,",
    isEnabled = false,
    scale = 1,
    config = {},
}

local onHeartPickupRule = {
    name = KnownRuleNames.OnHeartPickup,
    desc = "Picking up hearts",
    tags = "standardPointsRuleset,",
    isEnabled = false,
    scale = 1,
    config = {},
}

local onChestOpenRule = {
    name = KnownRuleNames.OnChestOpen,
    desc = "Opening chests",
    tags = "standardPointsRuleset,",
    isEnabled = false,
    scale = 1,
    config = {},
}

local onOrbPickupRule = {
    name = KnownRuleNames.OnOrbPickup,
    desc = "Picking up orbs",
    tags = "standardPointsRuleset,",
    isEnabled = false,
    scale = 1,
    config = {},
}

local onReachNewDepthRule = {
    name = KnownRuleNames.OnReachNewDepth,
    desc = "Reaching a new depth",
    tags = "",
    isEnabled = false,
    scale = 1,
    config = {},
}

-- RuleLibrary = {}

-- function RuleLibrary:new(o)
--     local this = o or {
--         rules = {
--             [KnownRuleNames.OnBasicEnemyKill] = onBasicEnemyKillRule,
--             [KnownRuleNames.OnReachEachHolyMountain] = onReachEachHolyMountainRule,
--             [KnownRuleNames.OnBossKill] = onBossKillRule,
--             [KnownRuleNames.OnHeartPickup] = onHeartPickupRule,
--             [KnownRuleNames.OnChestOpen] = onChestOpen,
--             [KnownRuleNames.OnReachNewDepth] = onReachNewDepthRule,
--         },
--     }
--     setmetatable(this,self)
--     self.__index = self
    
--     return this
-- end

RuleManager = {}

function RuleManager:new(o)
    local this = o or {
        rules = {
            [KnownRuleNames.OnBasicEnemyKill] = onBasicEnemyKillRule,
            [KnownRuleNames.OnReachEachHolyMountain] = onReachEachHolyMountainRule,
            [KnownRuleNames.OnBossKill] = onBossKillRule,
            [KnownRuleNames.OnHeartPickup] = onHeartPickupRule,
            [KnownRuleNames.OnChestOpen] = onChestOpenRule,
            [KnownRuleNames.OnOrbPickup] = onOrbPickupRule,
            [KnownRuleNames.OnReachNewDepth] = onReachNewDepthRule,
        },
    }
    setmetatable(this,self)
    self.__index = self
    
    function this:setIsEnabled(ruleName, isEnabled)
        local rule = self.rules[ruleName]
        if(rule == nil) then
            print("RULE_ERROR : Rule "..ruleName.." doesn't exist")
        end
        rule.isEnabled = isEnabled
    end

    function this:setConfig(ruleName, config)
        local rule = self.rules[ruleName]
        rule.config = config
    end

    function this:isRuleEnabled(ruleName)
        local rule = self.rules[ruleName]
        if(rule == nil) then 
            print("RULE_ERROR : Rule "..ruleName.." doesn't exist")
            return false
         end
        return rule.isEnabled
    end

    function this:enableTaggedRules(tagName)
        for name,rule in pairs(self.rules) do
            if(string.find(rule.tags,tagName) ~= nil) then
                if(rule == nil) then
                    print("RULE_ERROR : Rule "..ruleName.." doesn't exist")
                end
                rule.isEnabled = true
                print("RuleManager :"..rule.name..":"..tostring(rule.isEnabled))
            end
        end
        
    end

    function this:AddRuleFromLibrary(ruleName)
        return self.activeRules[ruleName] ~= nil
    end
    
    return this
end