dofile_once("data/scripts/lib/utilities.lua")
dofile_once("mods/scoita/files/scripts/libs/json.lua")
dofile_once("mods/scoita/files/scripts/utils.lua")
dofile_once("mods/scoita/files/scripts/rules/RuleManager.lua")
dofile_once("mods/scoita/files/scripts/events/EventManager.lua")
dofile_once("mods/scoita/files/scripts/ruleTriggers/RuleTriggerManager.lua")
dofile_once("mods/scoita/files/scripts/rules/knownRuleTypes.lua")
dofile_once("mods/scoita/files/scripts/scoreManager/scoreManagerInterface.lua")

dofile_once("mods/scoita/files/scripts/apiTypes.lua")

local scoreManagerUi = dofile_once("mods/scoita/files/scripts/scoreManager/scoreManagerUi.lua")
local scoreManagerApiClient = dofile_once("mods/scoita/files/scripts/scoreManager/ScoreManagerApiClient.lua")

ScoreManager = {}

function ScoreManager:new(o)
    local this = o or {
      eventBuffer = {},
      ruleManager = RuleManager:new(),
      eventManager = EventManager:new(),
      currentRun = {
        gameModeInstance = nil,
        isGameValidForRun = false,
        isRunActive = false,
        errorsPreventingRunFromStarting = {},
        ruleHelpers = {
          triggers = {},
          actions = {},
        },

      },
      leaderboardInfo = {
        currentLeaderboardPlayerInfo = {
          playerLeaderboardId = nil,
        },
        uiHelpers = {
          lastRunId = nil,
        },
      },
      stateTrackers = {
        Az = { 
          currentScore = 0,
          localHighScore = 0,
          logs = {}, 
          trackingHelpers = {
            biomesPreviouslyVisited = {},
            holyMountainVisitedYCoOrd = {},
            runTimer = {
              enabled = false,
              totalSecondsAllowed = 0
            },
          },
          ruleStates = {
            triggerStates = {},
            actionStates = {},
          }
        }
      },
      availableGameModeInfo = nil
    }
    setmetatable(this,self)
    self.__index = self

    this.eventManager:connectToScoreManager(this)


    function this:init()

      this.eventManager:init()

      if(ModSettingGet(KnownModSettingNames.PlayerLeaderboardId) == nil) then
        local playerId = tostring(Random(0,999999999))..tostring(Random(0,999999999))..tostring(Random(0,999999999));
        ModSettingSet(KnownModSettingNames.PlayerLeaderboardId, playerId)
      end

      this.leaderboardInfo.currentLeaderboardPlayerInfo.playerLeaderboardId = ModSettingGet(KnownModSettingNames.PlayerLeaderboardId)

      this:getAvailableGameModesFromServer()
    end


    function this:getStateTrackerForplayer()
      return self.stateTrackers.Az
    end


    function this:devSetGameMode(gameModeIdentifier)
      local gameModeRun = scoreManagerApiClient:getCurrentGameModeRunForGameModeInstance(gameModeIdentifier)
      local gameModeInstanceFull = scoreManagerApiClient:getGameModeRun("doesntMatter",gameModeRun._id);
      self.currentRun.gameModeInstance = gameModeInstanceFull
    end


    function this:onModInitialising(ModTextFileSetContentFunc, ModMagicNumbersFileAddFunc)
      local nextRunIsScoitaRun = ModSettingGet(KnownModSettingNames.NextRunIsScoitaRun)
      if(nextRunIsScoitaRun) then 
        local nextRunGameModeInstanceId = ModSettingGet(KnownModSettingNames.NextRunGameModeInstanceId)
        local gameModeInstanceFull = scoreManagerApiClient:getGameModeRun("doesntMatter",nextRunGameModeInstanceId);
        if(gameModeInstanceFull == nil) then
          GamePrintImportant("Couldn't load Scoita run :(")
          ModSettingRemove(KnownModSettingNames.NextRunIsScoitaRun)
          ModSettingRemove(KnownModSettingNames.NextRunGameModeInstanceId)
          do return end
        end

        self:setWorldSeed(gameModeInstanceFull.gameModeInfo.seed, ModTextFileSetContentFunc, ModMagicNumbersFileAddFunc)
        
        ModSettingRemove(KnownModSettingNames.NextRunIsScoitaRun)
        ModSettingRemove(KnownModSettingNames.NextRunGameModeInstanceId)

        self.currentRun.gameModeInstance = gameModeInstanceFull
      else
        ModTextFileSetContentFunc("mods/scoita/files/magic_numbers.xml", [[<MagicNumbers WORLD_SEED="0"/>]])  
        ModMagicNumbersFileAddFunc("mods/scoita/files/magic_numbers.xml")
      end
    end


    function this:setWorldSeed(seed, ModTextFileSetContentFunc, ModMagicNumbersFileAddFunc)
        ModTextFileSetContentFunc("mods/scoita/files/magic_numbers.xml", [[<MagicNumbers WORLD_SEED="]]..tostring(seed)..[["/>]])  
        ModMagicNumbersFileAddFunc("mods/scoita/files/magic_numbers.xml")
    end


    function this:getCurrentGameModeInstance()
      return self.currentRun.gameModeInstance
    end


    function this:getCurrentRun()
      return self.currentRun
    end


    function this:isOnScoitaRun()
      return self.currentRun.gameModeInstance ~= nil
    end


    function this:getPlayerLeaderboardId()
      return self.leaderboardInfo.currentLeaderboardPlayerInfo.playerLeaderboardId
    end


    function this:isRunActive()
      return self.currentRun.isRunActive
    end


    function this:onPlayerSpawned()
      if(this.currentRun.gameModeInstance ~= nil) then
        local isValid = self:checkIfCurrentRunIsValidForSelectedScoitaRun()
        if(isValid) then
          self.currentRun.isGameValidForRun = true
          self.currentRun.isRunActive = true

          this:initGameBasedOnGameMode(this.currentRun.gameModeInstance)
          GamePrintImportant(this.currentRun.gameModeInstance.gameMode.name.." run Started !")
        else
          scoreManagerUi:showGameModeIssuesUi()
        end
      end

    end


    function this:checkModsAgainstGameMode(gameModeidentifier, mods)

      if(__SCOITA__DEV__OVERRIDES__.ignoreModRequirements) then
        do return { success = true } end
      end

      return scoreManagerApiClient:checkModsAgainstGameModeRequirements(gameModeidentifier,mods)
    end


    function this:checkIfCurrentRunIsValidForSelectedScoitaRun()
      local hasErrors = false

      if(__SCOITA__DEV__OVERRIDES__.ignoreModRequirements) then
        do return true end
      end

      -- Scoita mod version
      local scoitaModVersion = ModSettingGet(KnownModSettingNames.Version)
      local scoitaVersionCheckResult = scoreManagerApiClient:getModVersionInfo(scoitaModVersion)

      if(not scoitaVersionCheckResult.isLatest) then
        local scoutaVersionMsg = "Scoita mod version outdated. Update to version "..scoitaVersionCheckResult.latestInfo.latestVersion
        hasErrors = true
        GamePrintImportant(scoutaVersionMsg)
        table.insert(self.currentRun.errorsPreventingRunFromStarting,scoutaVersionMsg)
      end
      
      -- gameMode
      local selectedGameModeInstance = this.currentRun.gameModeInstance
      local expectedGameMode = selectedGameModeInstance.gameMode.modeInfo.config.noitaGameModeType
      local currentRunContext = this:getCurrentRunContext();
      
      
      if(expectedGameMode ~= currentRunContext.gameMode) then
        hasErrors = true
        local wrongGameModeMessage = "You're on the wrong game mode for this run type !";
        table.insert(self.currentRun.errorsPreventingRunFromStarting,wrongGameModeMessage)
        table.insert(self.currentRun.errorsPreventingRunFromStarting,"Pick the run from the Leaderboard menu again and then")

        if(expectedGameMode == GameModes.Standard) then
          local msg = "Start a new game (Not a daily run)";
          table.insert(self.currentRun.errorsPreventingRunFromStarting,msg)

        elseif(expectedGameMode == GameModes.DailyRun) then
          local msg = "Start a new daily run !";
          table.insert(self.currentRun.errorsPreventingRunFromStarting,msg)
        else
          local msg = "Start the "..expectedGameMode.." game mode";
          GamePrintImportant(msg)
          table.insert(self.currentRun.errorsPreventingRunFromStarting,msg)
        end
      end

      -- mods
      -- Pass players mods up when requesting game modes 
      local modCheckResult = self:checkModsAgainstGameMode(selectedGameModeInstance.gameMode.identifier,currentRunContext.mods);
      if(not modCheckResult.success) then
        hasErrors = true
        
        self:printWarningsBasedOnModCheckResults(modCheckResult, true)
      end

      return not hasErrors
    end


    function this:printWarningsBasedOnModCheckResults(modCheckResult, insertErrorsToRunErrorList)
      
      local msg = "You have the wrong mods installed for this run!";
      GamePrintImportant(msg)
      if(insertErrorsToRunErrorList) then table.insert(self.currentRun.errorsPreventingRunFromStarting, msg) end

      if(tablelength(modCheckResult.additionalMods) > 0) then
        local msg = "Disable : "..json.encode(modCheckResult.additionalMods)
        GamePrintImportant(msg)
        if(insertErrorsToRunErrorList) then table.insert(self.currentRun.errorsPreventingRunFromStarting, msg) end
      end

      if(tablelength(modCheckResult.missingMods) > 0) then
        local msg = "Enable : "..json.encode(modCheckResult.missingMods);
        GamePrintImportant(msg)
        if(insertErrorsToRunErrorList) then table.insert(self.currentRun.errorsPreventingRunFromStarting, msg) end
      end
    end


    function this:initGameBasedOnGameMode(gameModeInstance)
      self.eventManager:initGameBasedOnGameMode(gameModeInstance)
    end


    function this:generateFreeRun(gameMode)
      return scoreManagerApiClient:getFreeRunForGameModeInstance(gameMode.identifier)
    end

    
    function this:setSelectedRunForNextGame(gameMode, runType, gameModeInstance)
      local modCheckResult = self:checkModsAgainstGameMode(gameMode.identifier,this:getCurrentRunContext().mods)
      if(not modCheckResult.success) then
        self:printWarningsBasedOnModCheckResults(modCheckResult, false)
        do end return
      else
        GamePrintImportant("Start a new game from menu to begin run")
      end

      local gameModeRun = nil
      if(gameModeInstance ~= nil) then
        gameModeRun = gameModeInstance
      else
        if(runType == KnownRunTypes.SingleAttempt) then
          gameModeRun = scoreManagerApiClient:getCurrentGameModeRunForGameModeInstance(gameMode.identifier)
        else
          gameModeRun = scoreManagerApiClient:getFreeRunForGameModeInstance(gameMode.identifier)
        end
      end
      
      ModSettingSet(KnownModSettingNames.NextRunIsScoitaRun, true)
      ModSettingSet(KnownModSettingNames.NextRunGameModeInstanceId, gameModeRun._id)
    end


    function this:getCurrentRunContext()
      return {
        seed = StatsGetValue("world_seed"),
        gameMode = GameIsDailyRun() and GameModes.DailyRun or GameModes.Standard,
        mods = ModGetActiveModIDs(),
      }
    end


    function this:getAvailableGameModesFromServer()
      this.availableGameModeInfo = scoreManagerApiClient:getAvailableGameModes()

      -- Grab some values from the server for faster iterations
      if(__SCOITA__DEV__OVERRIDES__.loadDevHelp) then
        __SCOITA__DEV__OVERRIDES__.devHelp = scoreManagerApiClient:getDevHelpObject()
      end
    end


    function this:getAvailableGameModes()
      return this.availableGameModeInfo
    end


    function this:checkScoreEvents()
      self.eventManager:checkScoreEvents()
    end


    function this:getActiveRunDurationInSeconds()
      return math.floor(GameGetFrameNum() / 60)
    end


    function this:getSecondsRemainingInRun()
      local secondsSinceStartedRun = this:getActiveRunDurationInSeconds();
      local toRet = this.stateTrackers.Az.trackingHelpers.runTimer.totalSecondsAllowed - secondsSinceStartedRun
      if(toRet < 0) then return 0 end
      return toRet
    end


    function this:getTotalSecondsForTimeLimitedRun()
      return this.stateTrackers.Az.trackingHelpers.runTimer.totalSecondsAllowed
    end


    function this:processBufferedScoreEvents()     
      self.eventManager:processBufferedScoreEvents()
    end


    function this:setScoreForPlayer(newScore)
      self.stateTrackers.Az.currentScore = newScore
    end


    function this:getScoreLogForPlayer()
      return self.stateTrackers.Az.logs
    end


    function this:getScoreForPlayer()
      return self.stateTrackers.Az.currentScore
    end


    function this:getLeaderboardInfo()
      return this.leaderboardInfo
    end


    function this:superMurderThePlayer()
      -- This is causing issues atm, disabled till I can remember why I did it
      --EntityInflictDamage(EntityGetWithTag("player_unit")[1],1000000, "DAMAGE_CURSE", "Run time over","DISINTEGRATED",0,0)
      EntityKill(EntityGetWithTag("player_unit")[1])
    end

    
    function this:renderScoreUi()
      scoreManagerUi:renderScoreUi(self, scoreManagerApiClient)
    end


    function this:onPlayerDeath()
      if(this.currentRun.isGameValidForRun) then
        local runSendResult = this:sendRunToLeaderboard();
        self.leaderboardInfo.uiHelpers.lastRunId = runSendResult.lastRunId;
        self.currentRun.isRunActive = false;
        
        scoreManagerUi:onLeaderboardClick()
        scoreManagerUi:goToPlayersHighestScoreOnLeaderboard()
        scoreManagerUi:showLeaderboard()

      else
        GamePrint("Score not uploaded as game setup wasn't valid for run")
      end

    end


    function this:sendRunToLeaderboard()
      local didPlayerCompleteScoitaRun = GameHasFlagRun( "scoita_run_completed" );
      local scoitaModVersion = ModSettingGet(KnownModSettingNames.Version)
      -- check if scoita managed game?
      local didCompleteRun = didPlayerCompleteScoitaRun

      local isRunAlwaysWonOnDeath = this.currentRun.gameModeInstance.gameMode.modeInfo.config.scoreRulesInfo.isRunAlwaysWonOnDeath
      if(isRunAlwaysWonOnDeath) then
        didCompleteRun = true
      end

      local runInfo = {
        playerInfo = {
          name = ModSettingGet(KnownModSettingNames.PlayerLeaderboardName),
          leaderboardId = this.leaderboardInfo.currentLeaderboardPlayerInfo.playerLeaderboardId,
        },
        additionalInfo = {
          ending_game_completed = ending_game_completed,
          didCompleteRun = didCompleteRun,
          context = this:getCurrentRunContext(),
          modVersions = {
            scoita = scoitaModVersion
          }
        },
        runTimeTakenInSeconds = this:getActiveRunDurationInSeconds(),
        score = this.stateTrackers.Az.currentScore,
        runModeContext = {
          gameModeIdentifier = this.currentRun.gameModeInstance.gameMode.identifier,
          runInstanceId = this.currentRun.gameModeInstance._id
        },
        stateTracker = this.stateTrackers.Az,
      }
      return scoreManagerApiClient:sendRunToLeaderboard(runInfo)
    end

    -- local world_entity_id = GameGetWorldStateEntity()
    -- local comp_worldstate = EntityGetFirstComponent( world_entity_id, "WorldStateComponent" )
    -- local flags = ComponentGetValue2(comp_worldstate, "flags")
    -- for i, v in ipairs(flags) do
    --   print(v)
    -- end
    
    return this
end

scoreManager = ScoreManager:new()