dofile_once("mods/scoita/files/scripts/libs/json.lua")
dofile_once("data/scripts/lib/utilities.lua")

local pollnet = dofile_once("mods/scoita/files/scripts/libs/pollnet/pollnet.lua")

local ScoreManagerApiClient = {}

local baseApiUrl = "http://scoita.uksouth.cloudapp.azure.com/noita/api/"

if(__SCOITA__DEV__OVERRIDES__.useLocalhost == true) then
  baseApiUrl = "http://localhost:8000/noita/api/"
end

local function poll_blocking(sock)
    while sock:poll() do
      if sock:last_message() then
        return sock:last_message()
      end
      pollnet.sleep_ms(20)
    end
    error("Socket closed?", sock:last_message())
    sock:close()
end

function ScoreManagerApiClient:new(o)
    local this = o or {
    }
    setmetatable(this,self)
    self.__index = self

    function this:init()
    end

    function this:getStandardDailyRunInfo(date)
        date = date or ""
        resp = self:getJsonResponse("standardDailyRunSeed/"..date)
        return resp
    end

    function this:getCurrentGameModeRunForGameModeInstance(gameModeIdentifier)
      resp = self:getJsonResponse("gameModes/"..gameModeIdentifier.."/currentRun")
      return resp
    end

    function this:getFreeRunForGameModeInstance(gameModeIdentifier)
      resp = self:getJsonResponse("gameModes/"..gameModeIdentifier.."/freeRun")
      return resp
    end    

    function this:getGameModeRunForGameModeInstance(gameModeIdentifier, date)
      resp = self:getJsonResponse("gameModes/"..gameModeIdentifier.."/date/"..date)
      return resp
    end

    function this:checkModsAgainstGameModeRequirements(gameModeIdentifier,mods)
      local body = {mods = mods}
      resp = self:getJsonPostResponse("gameModes/"..gameModeIdentifier.."/modCheck", json.encode(body))
      return resp
    end

    function this:getGameModeRun(gameModeIdentifier,gameModeInstanceIdentifier)
      resp = self:getJsonResponse("gameModes/"..gameModeIdentifier.."/"..gameModeInstanceIdentifier)
      return resp
    end

    function this:getLeaderboardForGameModeInstance(gameModeInstanceId, startPage)
      local page = startPage or 0
      resp = self:getJsonResponse("leaderboard/"..gameModeInstanceId.."/"..tostring(page))
      return resp
    end

    function this:getAvailableGameModes()
      resp = self:getJsonResponse("gameModes")
      return resp
    end

    function this:getModVersionInfo(currentModVersion)
      resp = self:getJsonResponse("modVersion/check/scoita/"..tostring(currentModVersion))
      return resp
    end

    function this:getPageOfPlayersHighestRun(gameModeInstanceId, playerId)
      resp = self:getJsonResponse("leaderboard/"..gameModeInstanceId.."/firstPageForPlayer/"..playerId)
      return resp
    end

    function this:hasPlayerAttemptedRun(gameModeInstanceId, playerId)
      resp = self:getJsonResponse("leaderboard/"..gameModeInstanceId.."/hasPlayerAttemptedRun/"..playerId)
      return resp
    end

    function this:getDevHelpObject()
      resp = self:getJsonResponse("devHelp")
      return resp
    end

    function this:sendRunToLeaderboard(context)
        resp = self:getJsonPostResponse("scoreRecord/run", json.encode(context))
        return resp
    end

    

    function this:getJsonResponse(url)
        local fullUrl = baseApiUrl..url
  
        if(__SCOITA__DEV__OVERRIDES__.logApiRequests) then
          print("GET", fullUrl)
        end
        local sock = pollnet.http_get(fullUrl, {
          ["User-Agent"] = "pollnet/" .. pollnet.VERSION
        })
        local status = poll_blocking(sock)
        --print("HTTP Status:", status)
        
        local headers = poll_blocking(sock)
        --print("HTTP Headers:")
        --print(headers)
        local body = poll_blocking(sock)
        --print(body)
        sock:close()
        return json.decode(body)
    end

    function this:getJsonPostResponse(url, jsonBody)
        local fullUrl = baseApiUrl..url
  
        if(__SCOITA__DEV__OVERRIDES__.logApiRequests) then
          print("GET", fullUrl)
        end

        local sock = pollnet.http_post(fullUrl, {
          ["User-Agent"] = "pollnet/" .. pollnet.VERSION,
          ["content-type"] = "application/json"
        },
        jsonBody
        )
        local status = poll_blocking(sock)
        --print("HTTP Status:", status)
        
        local headers = poll_blocking(sock)
        --print("HTTP Headers:")
        --print(headers)
        local body = poll_blocking(sock)
        --print("Body:")
        --print(body)
        sock:close()
        body = json.decode(body)
        return body
    end
    
    return this
end

local scoreManagerApiClient = ScoreManagerApiClient:new()

return scoreManagerApiClient