local date = dofile_once("mods/scoita/files/scripts/libs/luaDate/luaDate.lua")
dofile_once("mods/scoita/files/scripts/libs/json.lua")
dofile_once("mods/scoita/files/scripts/apiTypes.lua")
local sdl = dofile_once("mods/scoita/files/scripts/libs/sdl2_ffi/sdl2_ffi.lua")
local ffi = require"ffi"

local ScoreManagerUi = {}

local hoverStates = {}

local colors = {
  interesting = {
    r = 1, g = 1, b = 0, a = 1
  },
  interactive = {
    r = 1, g = 1, b = 0, a = 1
  },
  selected = {
    r = 1, g = 1, b = 0.6980392156862745, a = 1
  },
  veryBad = {
    r = 1, g = 0, b = 0, a = 1
  },
  conditionPassed = {
    r = 1, g = 1, b = 0, a = 0.5
  },
  subTitle = {
    r = 1, g = 1, b = 1, a = 0.8
  },
}

local showScoreInfo = false
local showLeaderboardMenu = false
local showrunContextInfo = false

LeaderboardModes = {
  None = "none",
  ViewLeaderboard = "ViewLeaderboard",
  FeaturedRun = "featuredRun",
}

local 

ButtonBackgrounds =
{
  Standard = "data/ui_gfx/decorations/9piece0_gray.png",
  Selected = "data/ui_gfx/decorations/9piece0.png",
}

local _id = 1
local function id()
  _id = _id + 1
  return _id
end

function ScoreManagerUi:new(o)
    local this = o or {
      scoreManager = nil,
      shouldShowGameModeIssues = false,
      leaderboardUiState = {
        mode = LeaderboardModes.None,
        modeSetupState = {
          gameMode = nil,
          hasPlayerAttemptedRun = false,
          freeRunInstances = {},
        },
        scoreboardState = {
          gameModeInstance = nil,
          currentDate = date(),
          scoreRunData = nil,
          currentPage = 0,
        },
        winChievementsState = {
          isExpanded = true,
        }
      }
    }

    setmetatable(this,self)
    self.__index = self

    function this:showGameModeIssuesUi()
        self.shouldShowGameModeIssues = true
    end

    function this:sendScoreToLeaderboard(run)
    end

    function this:showLeaderboard()
      showLeaderboardMenu = true;
    end

    function this:renderScoreUi(scoreManager, scoreManagerApiClient)
      _id = 1
      self.scoreManager = scoreManager
      self.scoreManagerApiClient = scoreManagerApiClient
      
      gui = gui or GuiCreate()
      GuiLayoutBeginLayer( gui )
      GuiStartFrame( gui )
    
      self:renderScoreSection(gui,560, 8)

      GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
      if(GuiButton(gui, id(), 480, 0, "Scoita menu")) then
        showLeaderboardMenu = not showLeaderboardMenu

        if(showLeaderboardMenu and self.leaderboardUiState.mode == LeaderboardModes.None) then
          if(self.scoreManager:isOnScoitaRun()) then
            self:onLeaderboardClick()
          else
            self:onFeaturedRunClick()
          end

        end

      end

      self:renderTimeSection(gui,560, 0)

      if(showScoreInfo) then
        self:renderScoreInfo()
      end

      if(showLeaderboardMenu and GameIsInventoryOpen() == false) then
        self:showLeaderboardMenu()
      end     

      if(self.shouldShowGameModeIssues) then
        this:renderRunStartIssuesUi(gui, 0,0)
      end

      
      self:renderWinConditionsAndAchievementsSection(gui, 340, 0)
      
      GuiLayoutEndLayer( gui )
    end

    function this:renderWinConditionsAndAchievementsSection(gui, baseX, baseY)
      local curY = baseY
      local triggers = self.scoreManager.eventManager:getTriggersForRun();
      local triggerStates = self.scoreManager.eventManager:getTriggerStatesForPlayer();
      local winConditionTriggers = {}
      local hasWinConditions = false
      local winConditionsMet = 0
      for id,trigger in pairs(triggers) do
        if(trigger:hasTag(ScoitaRuleTriggerTags.winCondition)) then
          hasWinConditions = true
          local triggerState = triggerStates[trigger.id]
          
          if(triggerState.state.isTriggered == true) then
            winConditionsMet = winConditionsMet + 1
          end

          table.insert(winConditionTriggers, trigger)
        end
      end

      if(hasWinConditions) then
        GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
        local isExpanded = self.leaderboardUiState.winChievementsState.isExpanded
        if(GuiButton( gui, id(), baseX, baseY, "Win Conditions ("..tostring(winConditionsMet).."/"..tablelength(winConditionTriggers)..")")) then
          self.leaderboardUiState.winChievementsState.isExpanded = not self.leaderboardUiState.winChievementsState.isExpanded
        end
        GuiTooltip(gui, "Click to show details", "")
        curY = curY + 10

        if(isExpanded and GameIsInventoryOpen() == false) then
          for id,trigger in pairs(winConditionTriggers) do
            local triggerState = triggerStates[trigger.id]
            local prefix = " - "
            if(triggerState.state.isTriggered == true) then
              GuiColorSetForNextWidget(gui, colors.conditionPassed.r,colors.conditionPassed.g,colors.conditionPassed.b,colors.conditionPassed.a)
              prefix = " / "
            end
            if(triggerState.state.isFailed == true) then
              GuiColorSetForNextWidget(gui, colors.veryBad.r,colors.veryBad.g,colors.veryBad.b,colors.veryBad.a)
            end
            
            GuiText( gui, baseX, curY, prefix..triggerState.description.oneLinerWithState)
            curY = curY + 10
          end
        end
      end
      

    end

    function this:renderScoreSection(gui, baseX, baseY)
      local score = self.scoreManager:getScoreForPlayer()

      if(self.scoreManager:isOnScoitaRun()) then
        local instance = self.scoreManager:getCurrentGameModeInstance()
        if(instance.gameMode.modeInfo.config.scoreRulesInfo.typeMeta == ScoitaScoreTypeMeta.time) then
          score = secondsToConsistentTimeString(score)
        end
      end
      
      
      local scoreLogClick = GuiButton(gui, id(), baseX, baseY, "Score")
      GuiTooltip(gui, "Show Score log", "")
      GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
      local scoreLog2Click = GuiButton(gui, id(), baseX + 25, baseY, score)
      GuiTooltip(gui, "Show Score log", "")
    
      if (scoreLogClick or scoreLog2Click) then
        showScoreInfo = not showScoreInfo
      end

    end

    function this:renderTimeSection(gui, baseX, baseY)
      if(self.scoreManager.stateTrackers.Az.trackingHelpers.runTimer.enabled and self.scoreManager:isRunActive()) then
        local secondsRemainingInRun = self.scoreManager:getSecondsRemainingInRun()
        local fullHrs,fullMins,fullSecs = getHoursMinutesSecondsFromSeconds(self.scoreManager:getTotalSecondsForTimeLimitedRun())
        local fullTime = tostring(fullHrs).." hr "..tostring(fullMins).." minutes and "..tostring(fullSecs).." seconds"

        local hrs,mins,secs = getHoursMinutesSecondsFromSeconds(secondsRemainingInRun)
        if(hrs >= 1) then
          GuiButton( gui, id(), baseX, baseY, "Time left "..padTimeElement(hrs)..":"..padTimeElement(mins))
        else
          GuiButton( gui, id(), baseX, baseY, "Time left "..padTimeElement(mins)..":"..padTimeElement(secs))  
        end
        
        GuiTooltip(gui, "This is a time-limited run", "You die when this runs out\r\n"..fullTime.."\r\nto get points or die tryin'")
      end
    end

    function this:renderRunStartIssuesUi(gui, baseX, baseY)
      local guiZDepthBase = 0;
      local baseX, baseY = 15, 50
      local currentRunInfo = self.scoreManager:getCurrentRun()

      GuiZSet(gui, guiZDepthBase)
      GuiImageNinePiece( gui, id(), baseX, baseY, 340, 100 )
      guiZDepthBase = guiZDepthBase -1
      GuiZSet(gui, guiZDepthBase)

      GuiColorSetForNextWidget(gui, 1,1,0,1)
      if (GuiButton( gui, id(), baseX + 160, baseY, "Ignore : I understand my run won't be recorded")) then
        self.shouldShowGameModeIssues = false
      end

      GuiColorSetForNextWidget(gui, 1,0,0,1)
      GuiText( gui, baseX + 5, baseY + 10, "WARNING !!!")
      
      GuiColorSetForNextWidget(gui, 1,0,0,1)
      GuiText( gui, baseX + 5, baseY + 20, "The run hasn't been configured correctly")

      GuiColorSetForNextWidget(gui, 1,0,0,1)
      GuiText( gui, baseX + 5, baseY + 30, "Your precious score won't be recorded unless you fix these issues:")
      

      local curY = baseY + 40
      for k,issue in pairs(currentRunInfo.errorsPreventingRunFromStarting) do
        GuiText( gui, baseX + 15, curY, issue)
        curY = curY + 10
      end
      
    end

    
    function this:setPage(page)
      self.leaderboardUiState.currentPage = page
    end

    function this:getCurrentRunContext()
      return self.scoreManager:getCurrentGameModeInstance()
    end

    function this:goToPlayersHighestScoreOnLeaderboard()
      local gameModeInstance = self.leaderboardUiState.scoreboardState.gameModeInstance or self:getCurrentRunContext()
      local highestRunRes = self.scoreManagerApiClient:getPageOfPlayersHighestRun(gameModeInstance._id,self.scoreManager:getPlayerLeaderboardId())
      if(highestRunRes.didFindPlayer == false) then
        do return end
      end
      local page = highestRunRes.didFindPlayer and highestRunRes.page or nil
      self.leaderboardUiState.scoreboardState.currentPage = page
      self:loadLeaderboardForGameModeInstance(gameModeInstance)
    end

    function this:renderScoreButton(gui, guiZDepthBase, x, y, width, height, text, isSelected, onSelected)

      local buttonImg = ButtonBackgrounds.Standard;
      if(isSelected) then
        buttonImg = ButtonBackgrounds.Selected;
      end

      GuiZSetForNextWidget(gui, guiZDepthBase -1)
      
      local btnBackId = id();
      
      GuiImageNinePiece(gui, btnBackId, x, y, width, height, 1, buttonImg)
      local modeClicked, rclicked, hovered = GuiGetPreviousWidgetInfo(gui)
      local mouseX, mouseY = InputGetMousePosOnScreen()

      
      if(hovered) then
        local screenX = MagicNumbersGetValue("VIRTUAL_RESOLUTION_X")
        local screenY = MagicNumbersGetValue("VIRTUAL_RESOLUTION_Y")
        -- HACK - InputGetMousePosOnScreen always returns values as if the res is 1280 x 720
        local screenToUiScaleX = 0.5
        local screenToUiScaleY = 0.5

        GuiLayoutBeginLayer(gui)
        GuiImage(gui, ScoitaModConstants.HardcodedUiIdOffset + 1, mouseX * screenToUiScaleX, mouseY * screenToUiScaleY, "mods/scoita/files/assets/ui/GameModeElementBackground.png", 0.000001, 1, 1)
        GuiLayoutEndLayer(gui)
      end

      if(hovered and hoverStates[btnBackId] == nil) then
        GamePlaySound("data/audio/Desktop/ui.bank", "ui/button_select", 0, 0)
        hoverStates[btnBackId] = true
      end

      if(hoverStates[btnBackId] ~= nil and not hovered) then
        hoverStates[btnBackId] = nil
      end

      if(modeClicked) then
        GamePlaySound("data/audio/Desktop/ui.bank", "ui/button_click", 0, 0)
        onSelected()
      end
      if(hovered) then
        GuiColorSetForNextWidget(gui, colors.selected.r,colors.selected.g,colors.selected.b,colors.selected.a)
      end
      if(isSelected) then
        GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
      end
      GuiZSetForNextWidget(gui, guiZDepthBase -2)
      GuiText( gui, x + 2, y, text)
    end
    
    function this:onModeChange()
      self.leaderboardUiState.currentPage = 0
    end

    function this:onFeaturedRunClick() 
      self:onModeChange()
      self.scoreManager:getAvailableGameModesFromServer()
      self.leaderboardUiState.mode = LeaderboardModes.FeaturedRun
    end

    function this:onLeaderboardClick()
      self:onModeChange()
      if(self.scoreManager:getCurrentGameModeInstance() ~= nil) then
        self.leaderboardUiState.scoreboardState.currentDate = date()
        local gameModeInstance = self.scoreManager:getCurrentGameModeInstance()
        self:loadLeaderboardForGameModeInstance(gameModeInstance)
      end
      self.leaderboardUiState.mode = LeaderboardModes.ViewLeaderboard

    end

    function this:showLeaderboardMenu()
      local baseX, baseY = 15, 155
      local sectionBaseX, sectionBaseX = 0, 0
      local guiZDepthBase = 0;
      GuiZSet(gui, guiZDepthBase)
      GuiImageNinePiece( gui, id(), baseX, baseY, 380, 200 )

      guiZDepthBase = guiZDepthBase -10
      GuiZSet(gui, guiZDepthBase)
      
      if(GuiButton( gui, id(), baseX + 360, baseY + 0, "Close")) then
        showLeaderboardMenu = false
      end

      local onFeaturedRunClickLocal = function() self:onFeaturedRunClick() end
      local onLeaderboardClickLocal = function() self:onLeaderboardClick() end

      local onCurrentRunClick = function() self:onLeaderboardClick() end

      if(self.scoreManager:isOnScoitaRun()) then
        local currentRun = "Current Run : "..self.scoreManager:getCurrentRun().gameModeInstance.gameMode.name
        GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
        if(GuiButton( gui, id(), baseX + 150, baseY + 5, currentRun)) then
          onLeaderboardClickLocal()
        end
      end
      

      self:renderScoreButton(gui, guiZDepthBase, baseX + 5, baseY + 5, 60, 10, "Featured Runs", self.leaderboardUiState.mode == LeaderboardModes.FeaturedRun, onFeaturedRunClickLocal)
      self:renderScoreButton(gui, guiZDepthBase, baseX + 75, baseY + 5, 60, 10, "Leaderboard", self.leaderboardUiState.mode == LeaderboardModes.ViewLeaderboard, onLeaderboardClickLocal)    
      
      local leaderboardInfo = self.scoreManager.getLeaderboardInfo()
      local leaderboardContext = leaderboardInfo.currentLeaderboardRunContext
      local currentLeaderboardInfo = leaderboardInfo.currentLeaderboardInfo
      
      baseY = baseY + 20
      if(self.leaderboardUiState.mode == LeaderboardModes.FeaturedRun) then
        self:renderGameModeMenu(gui,baseX, baseY, guiZDepthBase, leaderboardContext, currentLeaderboardInfo)
      elseif(self.leaderboardUiState.mode == LeaderboardModes.ViewLeaderboard) then 
        self:renderLeaderboardForGameModeInstance(gui,baseX, baseY, guiZDepthBase)
      end
      
    end

    function this:hasFreeRunBeenGeneratedForGameMode(gameMode) 
      return self.leaderboardUiState.modeSetupState.freeRunInstances[gameMode.identifier] ~= nil
    end

    function this:getGeneratedFreeRunForMode(gameMode)
      return self.leaderboardUiState.modeSetupState.freeRunInstances[gameMode.identifier]
    end

    function this:getOrGenerateFreeRun(gameMode)
        local generatedFreeRun = self.leaderboardUiState.modeSetupState.freeRunInstances[gameMode.identifier]
        local run = nil
        if(generatedFreeRun == nil) then
          run = self.scoreManager:generateFreeRun(gameMode)
          self.leaderboardUiState.modeSetupState.freeRunInstances[gameMode.identifier] = run
        else
          run = generatedFreeRun
        end

        return run
    end

    function this:onShareFreeRunFromInstance(gameModeInstance)
      local c_str = ffi.new("char[?]", #gameModeInstance._id + 1)
      ffi.copy(c_str, gameModeInstance._id)
      sdl.SDL_SetClipboardText(c_str)
    end

    function this:onShareFreeRunClick(gameMode)
      local gameModeInstance = self:getGeneratedFreeRunForMode(gameMode)
      local c_str = ffi.new("char[?]", #gameModeInstance._id + 1)
      ffi.copy(c_str, gameModeInstance._id)
      sdl.SDL_SetClipboardText(c_str)
    end

    function this:renderGameModeMenu(gui, baseX, baseY, guiZDepthBase, leaderboardContext)

      baseY = baseY + 5
      local gameModes = self.scoreManager:getAvailableGameModes().gameModes
      local curY = baseY
      
      local onPlayRunClick = function(gameMode, runType)

        local freeRunInstance = self.leaderboardUiState.modeSetupState.freeRunInstances[gameMode.identifier];
        if(runType == KnownRunTypes.FreeRun) then
          scoreManager:setSelectedRunForNextGame(gameMode, runType, freeRunInstance)
        else
          scoreManager:setSelectedRunForNextGame(gameMode, runType)
        end

        showLeaderboardMenu = false
      end

      local onGenerateFreeRunClick = function(gameMode)
        run = self.scoreManager:generateFreeRun(gameMode)
        self.leaderboardUiState.modeSetupState.freeRunInstances[gameMode.identifier] = run
      end

      local onLoadSharedRunClick = function()
        local char_ptr = sdl.SDL_GetClipboardText()
        local text = ffi.string(char_ptr)
        local gameMode = self.leaderboardUiState.modeSetupState.gameMode
        local gameModeInstance = self.scoreManagerApiClient:getGameModeRun(gameMode.identifier,text)
        if(gameModeInstance == nil) then
          print('Bad Code')
          do return end
        end
        
        self.leaderboardUiState.modeSetupState.gameMode = gameModeInstance.gameMode
        self.leaderboardUiState.modeSetupState.freeRunInstances[self.leaderboardUiState.modeSetupState.gameMode.identifier] = gameModeInstance
        --scoreManager:setSelectedRunForNextGame(gameModeInstance.gameMode, runType, gameModeInstance)
      end



      local onViewLeaderboardForGameModeClick = function(gameMode, runType)

        local gameModeRun = nil
        if(runType == KnownRunTypes.FreeRun) then
          gameModeRun = self:getOrGenerateFreeRun(gameMode)
        else
          gameModeRun = self.scoreManagerApiClient:getCurrentGameModeRunForGameModeInstance(gameMode.identifier)
        end

        self.leaderboardUiState.scoreboardState.currentDate = date()
        self.leaderboardUiState.scoreboardState.currentPage = 0
        
        self:loadLeaderboardForGameModeInstance(gameModeRun)
        self.leaderboardUiState.mode = LeaderboardModes.ViewLeaderboard
      end

      -- Game Mode section

      GuiZSet(gui, guiZDepthBase - 5)
      GuiBeginScrollContainer( gui, id(), baseX + 10, baseY, 150, 165 )
  
      curY = 0
      local hasPlayerAttemptedRun = false
      for k,gameMode in pairs(gameModes) do
        local shouldShowGameMode = gameMode.runInfo.shouldShowToPlayers
        if(shouldShowGameMode) then
          GuiZSet(gui, guiZDepthBase - 4)

          local gameModeBackground = "mods/scoita/files/assets/ui/GameModeElementBackground.png"
          if(self.leaderboardUiState.modeSetupState.gameMode ~= nil and self.leaderboardUiState.modeSetupState.gameMode.identifier == gameMode.identifier) then
            gameModeBackground = "mods/scoita/files/assets/ui/GameModeElementBackgroundSelected.png"
          end

          local btnBackId = id()
          GuiImage(gui, btnBackId, 0, curY, gameModeBackground, 1, 1, 1)
          local modeClicked, rclicked, hovered = GuiGetPreviousWidgetInfo(gui)

          if(hovered and hoverStates[btnBackId] == nil) then
            GamePlaySound("data/audio/Desktop/ui.bank", "ui/button_select", 0, 0)
            hoverStates[btnBackId] = true
          end
    
          if(hoverStates[btnBackId] ~= nil and not hovered) then
            hoverStates[btnBackId] = nil
          end

          if(modeClicked) then
            self.leaderboardUiState.modeSetupState.gameMode = gameMode
            local gameModeRun = self.scoreManagerApiClient:getCurrentGameModeRunForGameModeInstance(gameMode.identifier)
            self.leaderboardUiState.modeSetupState.hasPlayerAttemptedRun = self.scoreManagerApiClient:hasPlayerAttemptedRun(gameModeRun._id, self.scoreManager:getPlayerLeaderboardId()).hasExistingRun
            GamePlaySound("data/audio/Desktop/ui.bank", "ui/button_click", 0, 0)
          end
          
          GuiZSet(gui, guiZDepthBase - 10000)
          GuiImage(gui, id(), 10 + gameMode.logo.offset.x, curY + 10 + gameMode.logo.offset.y, gameMode.logo.path, 1, gameMode.logo.scale.x, gameMode.logo.scale.y)
          
          if(hovered) then
            GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
          end

          if(self.leaderboardUiState.modeSetupState.gameMode ~= nil and self.leaderboardUiState.modeSetupState.gameMode.identifier == gameMode.identifier) then
            GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
          end
          
          GuiText( gui, 40, curY + 5, gameMode.name)
          curY = curY + 40

        end
      end
      GuiEndScrollContainer( gui )

      -- Details section
      local baseXForDetailsSection = baseX + 180
      GuiZSet(gui, guiZDepthBase - 5)
      GuiBeginScrollContainer( gui, id(), baseXForDetailsSection, baseY, 180, 80 )
      if(self.leaderboardUiState.modeSetupState.gameMode ~= nil) then
        local curY = 0
        local selectedGameMode = self.leaderboardUiState.modeSetupState.gameMode;
        GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
        GuiText(gui, 0, curY, selectedGameMode.name)

        curY = curY + 10
        for dk,dLine in pairs(selectedGameMode.descriptionLines) do
          GuiZSet(gui, guiZDepthBase - 1)
          GuiText(gui, 10, curY, dLine)
          curY = curY + 10
        end

        curY = curY + 10

        local scoreRulesInfo = selectedGameMode.modeInfo.config.scoreRulesInfo
        local friendlyScoreCalcType = scoreRulesInfo.calculationType
        if(scoreRulesInfo.calculationType == ScoitaScoreCalculationType.distanceFromOrigin) then
          
          if(scoreRulesInfo.additionalConfig.measuredBy == ScoitaDistanceFromOriginMeasurement.down) then
            friendlyScoreCalcType = "Lowest point from surface"
          else
            friendlyScoreCalcType = "Highest point from surface"
          end

        elseif(scoreRulesInfo.calculationType == ScoitaScoreCalculationType.pointsSystem) then
          friendlyScoreCalcType = "Points system"
        elseif(scoreRulesInfo.calculationType == ScoitaScoreCalculationType.timeTaken) then
          friendlyScoreCalcType = "Time taken"
        end

        GuiText(gui, 0, curY, "Score based on : "..friendlyScoreCalcType)
        curY = curY + 10

        local isTimedRun = selectedGameMode.modeInfo.config.hasTimeLimit
        if(isTimedRun) then
          GuiText(gui, 0, curY, "Time limit "..secondsToConsistentTimeString(selectedGameMode.modeInfo.config.timeLimit))
          curY = curY + 10
        else
          GuiText(gui, 0, curY, "Time limit : None")
          curY = curY + 10
        end

        local modList = selectedGameMode.modeInfo.config.fullModStamp
        curY = curY + 5
        GuiText(gui, 0, curY, "Required Daily run modlist")
        GuiText(gui, 10, curY + 10, modList)
        -- for k,v in pairs(gameModeInstance.mods) do
        --   modList = modList.."\r\n  "..v
        -- end


      else
        GuiText(gui, 0, 0, "Select a game mode")
      end
      GuiEndScrollContainer( gui )

      local baseXForActionSection = baseX + 180
      local baseYForActionSection = baseY + 90
      if(self.leaderboardUiState.modeSetupState.gameMode ~= nil) then
          local selectedGameMode = self.leaderboardUiState.modeSetupState.gameMode;
        
          if(selectedGameMode.runInfo.cadence == ScoitaGameModeRotationCadence.daily) then
          
            GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
            GuiText( gui, baseXForActionSection, baseYForActionSection, "Daily run")
            GuiTooltip(gui,"Daily Run","Once chance per 24 hours\r\neveryone plays the same seed")
            GuiText( gui, baseXForActionSection + 38, baseYForActionSection, selectedGameMode.currentInstanceInfo.activeDate)    

            local curTime = os.time()
            local newRunTime = selectedGameMode.currentInstanceInfo.nextCadenceAt
            local timeRemainingToNextCadence = newRunTime - curTime
            local nextCadencePrint = ""
            local hastimeLeftToStartDailyRun = timeRemainingToNextCadence > 0
            
            GuiColorSetForNextWidget(gui, colors.subTitle.r,colors.subTitle.g,colors.subTitle.b,colors.subTitle.a)
            if(hastimeLeftToStartDailyRun) then
              nextCadencePrint = "(time left "..secondsToConsistentTimeString(timeRemainingToNextCadence)
            else
              nextCadencePrint = "Loading new run"
              self.scoreManager:getAvailableGameModesFromServer()
            end
            
            GuiText( gui, baseXForActionSection + 110, baseYForActionSection, nextCadencePrint..")")
            
            if(self.leaderboardUiState.modeSetupState.hasPlayerAttemptedRun) then
              GuiText( gui, baseXForActionSection + 10, baseYForActionSection + 15, "Completed")
            else
              self:renderScoreButton(gui, guiZDepthBase, baseXForActionSection + 10, baseYForActionSection + 15, 40, 10, "Play", false, function() onPlayRunClick(selectedGameMode, KnownRunTypes.SingleAttempt) end)
            end
            
            self:renderScoreButton(gui, guiZDepthBase, baseXForActionSection + 60, baseYForActionSection + 15, 70, 10, "View Leaderboard", false, function() onViewLeaderboardForGameModeClick(selectedGameMode) end)
          end

        GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
        local baseYForFreeRunActionSection = baseYForActionSection + 35;
        GuiText( gui, baseXForActionSection, baseYForFreeRunActionSection, "Free Run")
        GuiTooltip(gui,"Free Run","Play as much as you like\r\nseed is different everytime")

        if(self:hasFreeRunBeenGeneratedForGameMode(selectedGameMode)) then
          self:renderScoreButton(gui, guiZDepthBase, baseXForActionSection + 10, baseYForFreeRunActionSection + 15, 40, 10, "Play", false, function() onPlayRunClick(selectedGameMode, KnownRunTypes.FreeRun) end)
          self:renderScoreButton(gui, guiZDepthBase, baseXForActionSection + 60, baseYForFreeRunActionSection + 15, 70, 10, "View Leaderboard", false, function() onViewLeaderboardForGameModeClick(selectedGameMode, KnownRunTypes.FreeRun) end)
          self:renderScoreButton(gui, guiZDepthBase, baseXForActionSection + 100, baseYForFreeRunActionSection + 40, 60, 10, "Copy run code", false, function() self:onShareFreeRunClick(selectedGameMode) end)
          GuiTooltip(gui,"Copies the run id to your clipboard","Share with your friends to play the same run")
        else
          self:renderScoreButton(gui, guiZDepthBase, baseXForActionSection + 10, baseYForFreeRunActionSection + 15, 60, 10, "Generate Run", false, function() onGenerateFreeRunClick(selectedGameMode) end)
        end

        self:renderScoreButton(gui, guiZDepthBase, baseXForActionSection + 10, baseYForFreeRunActionSection + 35, 75, 10, "Load run from code", false, function() onLoadSharedRunClick() end)
        GuiTooltip(gui,"Load a run code from your clipboard","Use to play the same run with your friends")
      end

      --GuiEndScrollContainer( gui )

      
      --GuiText( gui, baseX + 0, baseY + 10, "A featured mod collection to run together !")
    end

    function this:renderLeaderboardForGameModeInstance(gui, baseX, baseY, guiZDepthBase)
      baseY = baseY + 5
      baseX = baseX + 5
      local gameModeInstance = self.leaderboardUiState.scoreboardState.gameModeInstance;
      local scoreRunData = self.leaderboardUiState.scoreboardState.scoreRunData;
      local curY = baseY

      if(gameModeInstance == nil) then
        GuiText( gui, baseX + 5, baseY, "No run selected.")
        GuiText( gui, baseX + 5, baseY + 10, "Click 'View Leaderboard' on a run in the Featured runs section")
        GuiText( gui, baseX + 5, baseY + 20, "Or start a new run")
        do return end
      end
      
      GuiText( gui, baseX, baseY, "Leaderboard :")
      GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
      GuiText( gui, baseX + 55, baseY, gameModeInstance.gameMode.name)
      curY = curY + 10

      local helpInfo = "What gives me score?\r\nScoring rules are simple...for now :)\r\n   Killing enemies\r\n   Collecting Hearts/Orbs/Chests\r\n   Getting to each Holy Mountain\r\nKilling (most) Bosses"
      GuiButton( gui, id(), baseX + 360, baseY, "Help")
      GuiTooltip(gui,"Global Leaderboard!",helpInfo)

      if(gameModeInstance.gameMode.runInfo.cadence == "daily" and gameModeInstance.gameModeInfo.runType == KnownRunTypes.SingleAttempt) then
        self:renderDailyRunMenu(gui, baseX, curY)
        curY = curY + 10
      end

      if(gameModeInstance.gameModeInfo.runType == KnownRunTypes.FreeRun) then
        self:renderFreeRunMenu(gui, baseX, curY, gameModeInstance)
        curY = curY + 15
      end
      
      self:renderScoreboardSection(baseX, curY,  gameModeInstance, scoreRunData)
      curY = curY + 20
    end

    function this:loadLeaderboardForGameModeInstance(gameModeInstance)
      local currentPage = self.leaderboardUiState.scoreboardState.currentPage
      self.leaderboardUiState.scoreboardState.gameModeInstance = gameModeInstance
      self.leaderboardUiState.scoreboardState.scoreRunData = self.scoreManagerApiClient:getLeaderboardForGameModeInstance(self.leaderboardUiState.scoreboardState.gameModeInstance._id, currentPage)
    end

    function this:renderDailyRunMenu(gui, baseX, baseY)
      local curDate = self.leaderboardUiState.scoreboardState.currentDate
      
      GuiText( gui, baseX + 0, baseY + 0, "Daily Run for")
      GuiColorSetForNextWidget(gui, colors.interactive.r,colors.interactive.g,colors.interactive.b, colors.interactive.a)
      if (GuiButton( gui, id(), baseX + 60, baseY + 0, "<-")) then
        self:changeLeaderboardDateAndReload(-1)  
      end
      GuiText( gui, baseX + 75, baseY + 0, curDate:fmt("%Y-%m-%d"))
      GuiColorSetForNextWidget(gui, colors.interactive.r,colors.interactive.g,colors.interactive.b, colors.interactive.a)
      if(GuiButton( gui, id(), baseX + 138, baseY + 0, "->")) then
          self:changeLeaderboardDateAndReload(1)
      end
    end

    function this:renderFreeRunMenu(gui, baseX, baseY, gameModeInstance)      
      GuiText( gui, baseX + 0, baseY + 5, "Free run Id :")
      GuiColorSetForNextWidget(gui, colors.interesting.r,colors.interesting.g,colors.interesting.b,colors.interesting.a)
      if(GuiButton( gui, id(), baseX + 55, baseY + 5, gameModeInstance._id)) then
        self:onShareFreeRunFromInstance(gameModeInstance)
      end
    end
    
    function this:changeLeaderboardDateAndReload(inc)
      local newDate = date(self.leaderboardUiState.scoreboardState.currentDate):adddays(inc)
      self.leaderboardUiState.scoreboardState.currentDate = newDate
      
      local gameModeRun = self.scoreManagerApiClient:getGameModeRunForGameModeInstance(self.leaderboardUiState.scoreboardState.gameModeInstance.gameMode.identifier, self.leaderboardUiState.scoreboardState.currentDate:fmt("%Y-%m-%d"))
      self:loadLeaderboardForGameModeInstance(gameModeRun)
    end

    function this:renderScoreboardHeader(baseX, baseY, gameModeInstance, scoreboardInfo)
      local scoreRulesInfo = gameModeInstance.gameMode.modeInfo.config.scoreRulesInfo
      
      if(scoreRulesInfo.typeMeta == ScoitaScoreTypeMeta.points) then
        self:renderScoreboardHeaderForPoints(baseX, baseY, gameModeInstance, scoreboardInfo)
      elseif(scoreRulesInfo.typeMeta == ScoitaScoreTypeMeta.distanceFromOrigin) then
        self:renderScoreboardHeaderForDistanceFromOrigin(baseX, baseY, gameModeInstance, scoreboardInfo)
      elseif(scoreRulesInfo.typeMeta == ScoitaScoreTypeMeta.time) then
        self:renderScoreboardHeaderForTime(baseX, baseY, gameModeInstance, scoreboardInfo)
      end

    end

    function this:renderScoreboardHeaderForPoints(baseX, baseY, gameModeInstance, scoreboardInfo)
      GuiText( gui, baseX + 0, baseY, "-------------------------------------------------------")
      GuiText( gui, baseX + 0, baseY + 10, "Rank | Name                      | Score                  | Time         | Won?   | Details")
      GuiText( gui, baseX + 0, baseY + 20, "--------------------------------------------------")
    end

    function this:renderScoreboardHeaderForDistanceFromOrigin(baseX, baseY, gameModeInstance, scoreboardInfo)
      local scoreRulesInfo = gameModeInstance.gameMode.modeInfo.config.scoreRulesInfo
      local scoreName = "Score"
      if(scoreRulesInfo.additionalConfig.measuredBy == ScoitaDistanceFromOriginMeasurement.up) then
        scoreName = "Height (meters)"
      else
        scoreName = "Depth (meters)"
      end

      GuiText( gui, baseX + 0, baseY, "-----------------------------------------------------")
      GuiText( gui, baseX + 0, baseY + 10, "Rank | Name                      |                          | Time            | Details")
      GuiText( gui, baseX + 125, baseY + 10, scoreName)
      GuiText( gui, baseX + 0, baseY + 20, "-----------------------------------------------------")
    end

    function this:renderScoreboardHeaderForTime(baseX, baseY, gameModeInstance, scoreboardInfo)
      GuiText( gui, baseX + 0, baseY, "-------------------------------------------------------")
      GuiText( gui, baseX + 0, baseY + 10, "Rank | Name                     | Time         | Won?   | Details")
      GuiText( gui, baseX + 0, baseY + 20, "--------------------------------------------------")
    end

    function self:renderScoreboardSection(baseX, baseY, gameModeInstance, scoreboardInfo)
      
      -- Divider
      this:renderScoreboardHeader(baseX, baseY, gameModeInstance, scoreboardInfo)
      
      local curY = baseY + 30

      -- Top Ten with pages?
      for k,v in pairs(scoreboardInfo.scoreRuns) do
        self:renderScoreRunElement(baseX, curY, gameModeInstance, v)
        curY = curY + 10
      end


      local pageUiOffset = 130
      if(GuiButton( gui, id(), baseX + 35, baseY + pageUiOffset + 5, "<-" )) then
        
        if(self.leaderboardUiState.scoreboardState.currentPage >= 1) then
          self.leaderboardUiState.scoreboardState.currentPage = self.leaderboardUiState.scoreboardState.currentPage - 1
        end

        self:loadLeaderboardForGameModeInstance(self.leaderboardUiState.scoreboardState.gameModeInstance)
      end

      if(GuiButton( gui, id(),baseX + 105, baseY + pageUiOffset + 5, "->" )) then

        if(self.leaderboardUiState.scoreboardState.currentPage < scoreboardInfo.scoreRunsMetaData.totalPages - 1) then
          self.leaderboardUiState.scoreboardState.currentPage = self.leaderboardUiState.scoreboardState.currentPage + 1
        end

        self:loadLeaderboardForGameModeInstance(self.leaderboardUiState.scoreboardState.gameModeInstance)
      end

      local totalPages = scoreboardInfo.scoreRunsMetaData.totalPages
      if(scoreboardInfo.scoreRunsMetaData.totalPages == 0) then totalPages = 1 end
      GuiText( gui, baseX + 50, baseY + pageUiOffset + 5, "Page "..scoreboardInfo.scoreRunsMetaData.currentPage.." of "..totalPages )

      if(GuiButton( gui, id(),baseX + 130, baseY + pageUiOffset + 5, "Show my highest score" )) then
        self:goToPlayersHighestScoreOnLeaderboard()
      end
      
    end

    function this:guiTextHighlightIfPlayer(entry,isLastRun)
      local currentPlayerLeaderboardId = self.scoreManager:getPlayerLeaderboardId()
      if(currentPlayerLeaderboardId == entry.playerInfo.leaderboardId) then
        GuiColorSetForNextWidget(gui, 1,1,0,1)
      end
      if(isLastRun) then
        GuiColorSetForNextWidget(gui, 0,1,1,1)
      end
    end

    function this:renderScoreRunElement(baseX, baseY, gameModeInstance, scoreRun)

      local scoreRulesInfo = gameModeInstance.gameMode.modeInfo.config.scoreRulesInfo
      
      if(scoreRulesInfo.typeMeta == ScoitaScoreTypeMeta.points) then
        self:renderScoreRunElementForPoints(baseX, baseY, gameModeInstance, scoreRun)
      elseif(scoreRulesInfo.typeMeta == ScoitaScoreTypeMeta.distanceFromOrigin) then
        self:renderScoreRunElementForDistanceFromOrigin(baseX, baseY, gameModeInstance, scoreRun)
      elseif(scoreRulesInfo.typeMeta == ScoitaScoreTypeMeta.time) then
        self:renderScoreRunElementTime(baseX, baseY, gameModeInstance, scoreRun)
      end

    end


    function this:renderScoreRunElementForPoints(baseX, baseY, gameModeInstance, scoreRun)
      local isLastRun = self.scoreManager.leaderboardInfo.uiHelpers.lastRunId == scoreRun.id
      local timeTaken = secondsToConsistentTimeString(scoreRun.runTimeTakenInSeconds);
      local didCompleteRun = false
      if(scoreRun.additionalInfo ~= nil) then
        didCompleteRun = scoreRun.additionalInfo.didCompleteRun or false
      end

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 0, baseY, tostring(scoreRun.rank))
      
      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiButton(gui, id(), baseX + 30,baseY, string.sub(scoreRun.playerInfo.name,0,20))
      GuiTooltip(gui,scoreRun.playerInfo.name,"")
      
      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 125, baseY, scoreRun.score)

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 210, baseY, timeTaken)

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 265, baseY, (didCompleteRun and "Yes" or "No"))

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiButton( gui, id(), baseX + 305, baseY, "View")
      GuiTooltip(gui,"Coming soon :)","")
    end

    function this:renderScoreRunElementForDistanceFromOrigin(baseX, baseY, gameModeInstance, scoreRun)
      local isLastRun = self.scoreManager.leaderboardInfo.uiHelpers.lastRunId == scoreRun.id
      local timeTaken = secondsToConsistentTimeString(scoreRun.runTimeTakenInSeconds);

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 0, baseY, tostring(scoreRun.rank))
      
      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiButton(gui, id(), baseX + 30,baseY, string.sub(scoreRun.playerInfo.name,0,20))
      GuiTooltip(gui,scoreRun.playerInfo.name,"")
      
      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 125, baseY, scoreRun.score)

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 210, baseY, timeTaken)

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiButton( gui, id(), baseX + 270, baseY, "View")
      GuiTooltip(gui,"Coming soon :)","")
    end

    function this:renderScoreRunElementTime(baseX, baseY, gameModeInstance, scoreRun)
      local isLastRun = self.scoreManager.leaderboardInfo.uiHelpers.lastRunId == scoreRun.id
      local timeTaken = secondsToConsistentTimeString(scoreRun.runTimeTakenInSeconds);
      local didCompleteRun = false

      if(scoreRun.additionalInfo ~= nil) then
        didCompleteRun = scoreRun.additionalInfo.didCompleteRun or false
      end

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 0, baseY, tostring(scoreRun.rank))
      
      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiButton(gui, id(), baseX + 30,baseY, string.sub(scoreRun.playerInfo.name,0,20))
      GuiTooltip(gui,scoreRun.playerInfo.name,"")

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 125, baseY, timeTaken)

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiText( gui, baseX + 175, baseY, (didCompleteRun and "Yes" or "No"))

      self:guiTextHighlightIfPlayer(scoreRun, isLastRun)
      GuiButton( gui, id(), baseX + 270, baseY, "View")
      GuiTooltip(gui,"Coming soon :)","")
    end


    function this:renderScoreInfo()
      local guiZDepthBase = -100;
      local baseX, baseY = 280, 100
      GuiZSet(gui, guiZDepthBase -1)
      GuiText( gui, baseX, baseY, "Score Log")
      if(GuiButton(gui, id(), baseX + 280, baseY, "Close")) then
        showScoreInfo = false
      end

      GuiZSet(gui, guiZDepthBase)
      GuiImageNinePiece( gui, id(), baseX, baseY, 304, 10 )
      GuiBeginScrollContainer( gui, id(), baseX, baseY + 13, 300, 100 )
      GuiLayoutBeginVertical( gui, 0, 0 )
  
      latestLog = self.scoreManager:getScoreLogForPlayer();
      scoreLogLength = tablelength(latestLog)
  
      for i=scoreLogLength,1,-1  do 
        logEntry = latestLog[i]
        GuiText( gui, 0, 0, logEntry.desc .." : "..logEntry.points.."pts")
      end
      GuiLayoutEnd( gui )
      GuiEndScrollContainer( gui )
    end
    
    return this
end

local scoreManagerUi = ScoreManagerUi:new()

return scoreManagerUi