__SCOITA__DEV__OVERRIDES__ = {
	setGameMode = nil,
	ignoreModRequirements = false,
	useLocalhost = false,
  loadDevHelp = false,
  logApiRequests = false,
}

dofile_once("mods/scoita/files/scripts/devOverrides.lua")

function tablelength(T)
  local count = 0
  for _ in pairs(T) do count = count + 1 end
  return count
end

local modName = "scoita"

KnownModSettingNames = {
	LocalHighScore = modName..".localHighScore",
	Version = modName.."._version",
	PlayerLeaderboardId = modName..".playerLeaderboardId",
  PlayerLeaderboardName = modName..".playerLeaderboardName",
  NextRunGameModeInstanceId = modName..".nextRunGameModeInstanceId",
  NextRunIsScoitaRun = modName..".nextRunIsScoitaRun",
}

KnownRunTypes = {
  SingleAttempt = "singleAttempt",
  FreeRun = "freeRun",
}

ScoitaModConstants = {
  PointsScale = 10,
  LowestKillPoints = 10,
  HardcodedUiIdOffset = 7000
}

GameModes = {
  Standard = "standard",
  DailyRun = "dailyRun",
}

KnownGlobals = {
  ScoreEventBuffer = "scoita_score_event_buffer",
}

function getPlayersCurrentBiome()
  local player = EntityGetWithTag("player_unit")[1]
	local x, y = EntityGetTransform(player)
	local curBiome = BiomeMapGetName( x, y )
  return curBiome
end

function IsPlayerPolymorphed() -- returns bool, entityId/nil
  local polymorphed_entities = EntityGetWithTag("polymorphed")
  if (polymorphed_entities ~= nil) then
      for _, entity_id in ipairs(polymorphed_entities) do
          local is_player = false
          local game_stats_comp = EntityGetFirstComponent(entity_id, "GameStatsComponent")
          if (game_stats_comp ~= nil) then 
              is_player = ComponentGetValue2(game_stats_comp, "is_player")
          end
          if (is_player) then
              return true, entity_id
          end
      end
      return false, nil
  end
end

function getPlayerEntity()
	local players = EntityGetWithTag("player_unit")
	if #players == 0 then return end

	return players[1]
end

function getHoursMinutesSecondsFromSeconds(totalSeconds)
  local hours = math.floor(totalSeconds / 60 / 60);
  local minutesRemaining = math.floor((totalSeconds - (hours * 60 * 60)) / 60);
  local secondsRemaining = math.floor(totalSeconds % 60);

  return hours, minutesRemaining, secondsRemaining
end

function shallowCopyTable(orig)
  local orig_type = type(orig)
  local copy
  if orig_type == 'table' then
      copy = {}
      for orig_key, orig_value in pairs(orig) do
          copy[orig_key] = orig_value
      end
  else -- number, string, boolean, etc
      copy = orig
  end
  return copy
end

function stringToDate(date)
  
  p="%a+, (%d+) (%a+) (%d+) (%d+):(%d+):(%d+) GMT"
  day,month,year,hour,min,sec=date:match(p)
  MON={Jan=1,Feb=2,Mar=3,Apr=4,May=5,Jun=6,Jul=7,Aug=8,Sep=9,Oct=10,Nov=11,Dec=12}
  month=MON[month]
  offset=os.time()-os.time(os.date("!*t"))
  return os.time({day=day,month=month,year=year,hour=hour,min=min,sec=sec})
end

function secondsToConsistentTimeString(seconds)
  local fullHrs,fullMins,fullSecs = getHoursMinutesSecondsFromSeconds(seconds)
  return padTimeElement(fullHrs)..":"..padTimeElement(fullMins)..":"..padTimeElement(fullSecs)
  
end

function padTimeElement(elem)  
  local elemStr = tostring(elem)
  if(elem < 10) then elemStr = "0"..elemStr end
  return elemStr
  
end