dofile_once("data/scripts/lib/utilities.lua")
dofile_once("mods/scoita/files/scripts/libs/json.lua")
dofile_once("mods/scoita/files/scripts/utils.lua")

dofile_once("mods/scoita/files/scripts/scoreManager/ScoreManager.lua")



function setRunInfoIfScoitaRun()
	if(__SCOITA__DEV__OVERRIDES__.setGameMode ~= nil) then
		scoreManager:devSetGameMode(__SCOITA__DEV__OVERRIDES__.setGameMode)
	end
	scoreManager:onModInitialising(ModTextFileSetContentFunc, ModMagicNumbersFileAddFunc)
end


function applyModfileExtensions()

	local bossOnDeathTemplate = "mods/scoita/files/scripts/dataScriptExtensions/onDeathTemplate.lua";
	local bosses = { 
		alchemist = {
			name = "Alchemist",
			baseGameFile = "data/entities/animals/boss_alchemist/death.lua",
			targetFile = bossOnDeathTemplate,
			points = -1,
			scoreBasedOnHealth = true
		},
		centipede = {
			name = "centipede",
			baseGameFile = "data/entities/animals/boss_centipede/death_check.lua",
			targetFile = bossOnDeathTemplate,
			points = 500,
			scoreBasedOnHealth = false
		},
		fish = {
			name = "fish",
			baseGameFile = "data/entities/animals/boss_fish/death.lua",
			targetFile = bossOnDeathTemplate,
			points = -1,
			scoreBasedOnHealth = true
		},
		ghost = {
			name = "ghost",
			baseGameFile = "data/entities/animals/boss_ghost/death.lua",
			targetFile = bossOnDeathTemplate,
			points = -1,
			scoreBasedOnHealth = true
		},
		limbs = {
			name = "limbs",
			baseGameFile = "data/entities/animals/boss_limbs/boss_limbs_death.lua",
			targetFile = bossOnDeathTemplate,
			points = -1,
			scoreBasedOnHealth = true
		},
		meat = {
			name = "meat",
			baseGameFile = "data/entities/animals/boss_meat/death.lua",
			targetFile = bossOnDeathTemplate,
			points = -1,
			scoreBasedOnHealth = true
		},
		pit = {
			name = "pit",
			baseGameFile = "data/entities/animals/boss_pit/boss_pit_death.lua",
			targetFile = bossOnDeathTemplate,
			points = -1,
			scoreBasedOnHealth = true
		},
		robot = {
			name = "robot",
			baseGameFile = "data/entities/animals/boss_robot/death.lua",
			targetFile = bossOnDeathTemplate,
			points = -1,
			scoreBasedOnHealth = true
		},
		spirit = {
			name = "spirit",
			baseGameFile = "data/entities/animals/boss_spirit/death.lua",
			targetFile = bossOnDeathTemplate,
			points = -1,
			scoreBasedOnHealth = true
		},
		wizard = {
			name = "wizard",
			baseGameFile = "data/entities/animals/boss_wizard/death.lua",
			targetFile = bossOnDeathTemplate,
			points = -1,
			scoreBasedOnHealth = true
		},
	}

	for k,v in pairs(bosses) do

		local deathTemplate = ModTextFileGetContent("mods/scoita/files/scripts/dataScriptExtensions/onDeathTemplate.lua")
		local deathForBoss = deathTemplate
		deathForBoss = string.gsub(deathForBoss, "BOSSNAME", v.name)
		if(v.scoreBasedOnHealth) then
			deathForBoss = string.gsub(deathForBoss, "POINTSFORBOSS", "amount")
		else
			deathForBoss = string.gsub(deathForBoss, "POINTSFORBOSS", v.points)
		end
		
		local generatedScriptName = "mods/scoita/scripts/generated/"..v.name.."Death.lua";

		ModTextFileSetContent(generatedScriptName, deathForBoss)
		ModLuaFileAppend( v.baseGameFile, generatedScriptName)
	end


	ModLuaFileAppend( "data/scripts/items/heart.lua", "mods/scoita/files/scripts/dataScriptExtensions/heart.lua" )
	ModLuaFileAppend( "data/scripts/items/chest_random.lua", "mods/scoita/files/scripts/dataScriptExtensions/chest_random.lua" )
	ModLuaFileAppend( "data/scripts/items/orb_pickup.lua", "mods/scoita/files/scripts/dataScriptExtensions/orb_pickup.lua" )
end




function OnModInit()

	ModTextFileSetContentFunc = ModTextFileSetContent;
	ModMagicNumbersFileAddFunc = ModMagicNumbersFileAdd;
	
	setRunInfoIfScoitaRun()
	applyModfileExtensions()	
end


-- all functions below are optional and can be left out

--[[



function OnModPostInit()
	print("Mod - OnModPostInit()") -- Then this is called for all mods
end

function OnPlayerSpawned( player_entity ) -- This runs when player entity has been created
	GamePrint( "OnPlayerSpawned() - Player entity id: " .. tostring(player_entity) )
end

function OnWorldInitialized() -- This is called once the game world is initialized. Doesn't ensure any world chunks actually exist. Use OnPlayerSpawned to ensure the chunks around player have been loaded or created.
	GamePrint( "OnWorldInitialized() " .. tostring(GameGetFrameNum()) )
end

function OnWorldPostUpdate() -- This is called every time the game has finished updating the world
	GamePrint( "Post-update hook " .. tostring(GameGetFrameNum()) )
end

]]--

function OnPlayerSpawned( player_entity ) -- This runs when player entity has been created
	if(ModSettingGet(KnownModSettingNames.PlayerLeaderboardName) == "Anonymous Noita") then
		GamePrintImportant("Set your leaderboard name in the scoita mod settings!")
	end

	scoreManager:onPlayerSpawned()
end


function OnModPreInit()

end

function OnPlayerDied(playerEntity)
	scoreManager:onPlayerDeath()
end

function OnWorldInitialized() -- This is called once the game world is initialized. Doesn't ensure any world chunks actually exist. Use OnPlayerSpawned to ensure the chunks around player have been loaded or created.

	GlobalsSetValue("scoita_log","[]")
	scoreManager:init()
end

function OnMagicNumbersAndWorldSeedInitialized() -- this is the last point where the Mod* API is available. after this materials.xml will be loaded.
	local x = ProceduralRandom(0,0)
	print( "===================================== random " .. tostring(x) )
end

function OnWorldPreUpdate() -- This is called every time the game is about to start updating the world
	scoreManager:renderScoreUi();
end

function OnWorldPostUpdate() -- This is called every time the game has finished updating the world
	scoreManager:processBufferedScoreEvents()
	scoreManager:checkScoreEvents()

	--checkIfPlayerIsInNewBiome()
end

function checkIfPlayerIsInNewBiome()
	local player = EntityGetWithTag("player_unit")[1]
	local x, y = EntityGetTransform(player)
	local currbiome = BiomeMapGetName( x, y )
	for _,i in ipairs(biome_list) do
		inc = inc + 1
		if i == tostring(currbiome) then
			list_of_tiles_done,biome_list,biome_list_index = update_globals(list_of_tiles_done,biome_list,biome_list_index,"biome_list",inc)
			GamePrint("Visted ".. i .. " biome")
		end
	end
end

local replacement = nil
local fileContents = nil
-- adding customized listeners to base enemy entity definition file
fileContents = ModTextFileGetContent("data/entities/base_humanoid.xml")
replacement = [[
	</ItemChestComponent>
	<LuaComponent
		execute_every_n_frame="25"
	    script_death="mods/scoita/files/scripts/onDeath.lua"
	    remove_after_executed="1"
		execute_times="1">
	</LuaComponent>
]]
fileContents = string.gsub(fileContents, "</ItemChestComponent>", replacement)
ModTextFileSetContent("data/entities/base_humanoid.xml", fileContents )
fileContents = nil
replacement = nil


-- -- This code runs when all mods' filesystems are registered
-- ModLuaFileAppend( "data/scripts/gun/gun_actions.lua", "mods/example/files/actions.lua" ) -- Basically dofile("mods/example/files/actions.lua") will appear at the end of gun_actions.lua
-- ModMagicNumbersFileAdd( "mods/example/files/magic_numbers.xml" ) -- Will override some magic numbers using the specified file

-- -- see tools_modding/noita_fmod_project/ for the project that defines the audio events used in this mod
-- ModRegisterAudioEventMappings( "mods/example/files/audio_events.txt" ) -- Use this to register custom fmod events. Event mapping files can be generated via File -> Export GUIDs in FMOD Studio.
-- ModRegisterMusicBank( "mods/example/files/music.bank" ) -- Use this to register custom banks in the music system. ModRegisterAudioEventMappings also needs to be called to make the game recognize the events.

-- ModMaterialsFileAdd( "mods/example/files/materials_rainbow.xml" ) -- Adds a new 'rainbow' material to materials
-- ModLuaFileAppend( "data/scripts/items/potion.lua", "mods/example/files/potion_appends.lua" )

-- -- make coalmine play music from the new bank
-- local coal_xml = ModTextFileGetContent( "data/biome/coalmine.xml" )
-- coal_xml = coal_xml:gsub([[audio_music_2="coalmine"]], [[audio_music_2="thebiome"]] )
-- ModTextFileSetContent( "data/biome/coalmine.xml", coal_xml )


-- if ModImageMakeEditable ~= nil then -- needed to avoid error if this file is hotloaded after init
-- 	-- make the player sprite green
-- 	local t1 = GameGetRealWorldTimeSinceStarted();

-- 	local recolor = function( filename)
-- 		local id,w,h = ModImageMakeEditable( filename, 0, 0 )
-- 		for y=0,h do 
-- 			for x=0,w do
-- 				local c = ModImageGetPixel( id, x, y )
-- 	    		local r,g,b,a = color_abgr_split( c )
-- 	    		r = r / 2
-- 	    		b = b / 2
-- 	    		c = color_abgr_merge(r,g,b,a)
-- 				ModImageSetPixel( id, x, y, c )
-- 			end
-- 		end
-- 	end

-- 	-- using different slots here to test the feature works, 
-- 	-- but you could use the same slot when editing images independent of each other like this
-- 	recolor( "data/enemies_gfx/player.png" )
-- 	recolor( "data/enemies_gfx/player_arm.png" )
-- 	recolor( "data/enemies_gfx/player_arm_no_item.png" )

-- 	t1 = GameGetRealWorldTimeSinceStarted() -t1
-- 	print("ModImage stuff took " .. t1 .. " seconds")

-- 	local who_edited = ModImageWhoSetContent( "data/enemies_gfx/player.png" )
-- 	local ok = " - fail"
-- 	if who_edited == "example" then ok = " - ok" end

-- 	print( "Who edited? " .. who_edited .. ok )
-- end